/*******************************************************************************************************
    Search for classical Capacity of Quantum Channels(SeCQC) Version 1.0, a program that can calculate the maximum accessible information 
    and the corresponding optimal statistical operators as well as POVMs given a certain quantum channel.

    Copyright (C) 2010 J.W. Shang, K.L. Lee and B.-G. Englert.

    This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License Version 3 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR PARTICULAR PURPOSE.  
    See the GNU General Public License for more details.

    You can refer to http://www.gnu.org/licenses/ for the GNU General Public License.
	
    Detail manual for the program could be downloaded from http://www.quantumlah.org/publications/software/SeCQC/.
    
   	The conjugate-gradient method, golden section search method as well as LU decomposition are performed following the algorithm given in 'Numerical Recipe in C', 2nd Ed. 
    This edition is available online at http://www.nrbook.com/c/.
    
    This reference will be referred to as "NRC" in the comments of the following codes.
********************************************************************************************************/
//******************************************************************************************************/
//	Compiled under Windows XP Service Pack 3, with wxWidgets 2.8.10, MinGW 5.1.6, MSYS 1.0.11
//
//	Compile with the following command
//		g++ MI.cpp `wx-config --libs` `wx-config --cxxflags` -o SeCQC
//	in MSYS shell.
//
//  If you face problems running the program in a Linux environment, try
//      export LD LIBRARY PATH=/usr/local/lib
//  in command window.
//
//******************************************************************************************************/

//--------------------main function to calculate the mutual information--------------------------------//

#define TINY 1.0e-4

class MutInf{
	//------------------------------------------------------------------------------
	//	variable declarations
	//------------------------------------------------------------------------------
	public:
		MutInf(int d, int n, int j, int k, int m, int c);		
		void calcMutualInformation(); //main function to calculate the mutual information MI
				
		//variables
		int rhoDimension, povmDimension, directGradChance, maxDimen, maxMatrix;
		double MI, new_MI, alpha, epsilon, phi;
		double tol_MI, tol_alpha, tol_epsilon, tol_S;
		string output;
		vector< vector<double> > prob_table;
						
		//various functions dealing with actual calculations
		double calc_beta(vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		void calc_delta_A_dagger_CG(double, int, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&);		
		void calc_grad_MI(vector< vector<dcomp> >&, vector< vector< vector<dcomp> > >&, vector< vector<dcomp> >&);
		void calc_L_inverse(vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		inline double calc_MI(double some_alpha, double some_epsilon, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);		
		double calc_MI_from_prob_table(vector< vector<double> >&);	
		void calc_new_A_dagger(double some_alpha, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&);
        void calc_new_rho(double, double, vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);		
		void calc_prob_table(vector< vector<dcomp> >&, vector< vector< vector<dcomp> > >&);
		void calc_R(vector< vector< vector<dcomp> > >&, vector< vector< vector <dcomp> > >&);
		void calc_T(vector< vector< vector<dcomp> > >&, vector< vector<dcomp> >&);
		void find_alpha_and_epsilon(vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);
		void gen_A_dagger(vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		void gen_rho(vector< vector< vector<dcomp> > >&);
		void rho_after_Capacitor(vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);
		void povm_after_Capacitor(vector< vector< vector<dcomp> > >&, vector< vector<dcomp> >&);
		
        //functions for interaction with GUI
		int getK();
		int getJ();
		int getM();
		int getC();
		vector <vector <dcomp> > getPOVM(int X);
		vector <vector <dcomp> > getrho(int X);
		vector <vector <dcomp> > getCap(int X, int Y);
		void setrhoElement(int X, int Y, int Z, double d, int ri);
		void setPOVMElement(int X, int Y, int Z, double d, int ri);
		void setCapElement(int L, int X, int Y, int Z, double h, int ri);
		void UpdatePOVMDimension(int k, int n);
		void UpdaterhoDimension(int j, int n);
        void UpdateCapDimension(int c, int m, int d);			
		
		//functions initializing 1D, 2D, 3D and 4D matrices
		vector<dcomp> InitVector(int X);
		vector< vector<dcomp> > Init2DMatrix(int X, int Y);		
		vector< vector<double> > Init2DRealMatrix(int X, int Y);
		vector< vector< vector<dcomp> > > Init3DMatrix(int X, int Y, int Z);
		vector< vector< vector< vector<dcomp> > > > Init4DMatrix(int X, int Y, int Z, int D);
		
		//functions multiplying scalars, vectors and matrices
		dcomp inner_product(vector<dcomp>&, vector<dcomp>&);
		double LOG2(double);
		void matrix_add_matrix(vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		void matrix_times_matrix(vector< vector<dcomp> >&, vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		void matrix_times_vector(vector<dcomp>&, vector< vector<dcomp> >&, vector<dcomp>&);
		void outer_product_add(vector< vector<dcomp> >&, vector<dcomp>&, vector<dcomp>&);
		void scalar_times_vector(vector<dcomp>&, double, vector<dcomp>&);
		void scalar_times_matrix(vector< vector<dcomp> >&, double, vector< vector<dcomp> >&);
		void complex_times_matrix(vector< vector<dcomp> >&, dcomp, vector< vector<dcomp> >&);
		void swap2D(vector< vector<dcomp> >&, vector< vector<dcomp> >&);
		void swap3D(vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);
		void transpose_conjugate(vector< vector<dcomp> >&);
		dcomp cal_trace(vector< vector<dcomp> >&);
		double cal_modulus(dcomp);
	    void cal_kroneckerProduct(vector< vector< vector<dcomp> > >&, vector< vector< vector<dcomp> > >&);
	    int cal_power(int, int);
	
	private:
		int D, N, J, K, M, C;
		dcomp zero;
		vector<dcomp> rho2; /* 3rd dimension of rho */
		vector< vector<dcomp> > rho1; /* 2nd dimension of rho */
		vector< vector< vector<dcomp> > > rho; /* 1st dimension of rho. Dimension of rho is J x N x N */
		vector<dcomp> povm2; /* 3rd dimension of povm */
		vector< vector<dcomp> > povm1; /* 2nd dimension of povm */
		vector< vector< vector<dcomp> > > povm; /* 1st dimension of povm. Dimension of povm is K x N x N */	
		vector<dcomp> Cap3; /* 4th dimension of Cap */
		vector< vector<dcomp> > Cap2; /* 3rd dimension of Cap */
		vector< vector< vector<dcomp> > > Cap1; /* 2nd dimension of Cap. Dimension of Cap is C x M x N x N */
        vector< vector< vector< vector<dcomp> > > > Cap; /* 1st dimension of Cap. Dimension of Cap is C x M x N x N */
};

MutInf::MutInf(int d, int n, int j, int k, int m, int c)
	   :tol_MI(1.0e-15),tol_alpha(1E-13),tol_epsilon(1E-13),tol_S(1E-15),D(d),N(n),J(j),K(k),M(m),C(c),zero(0.0,0.0),rhoDimension(n),povmDimension(n),output("output.txt"),directGradChance(0)
{
 /* reserve enough memory space for rho and povm such as to avoid reallocation bug */
 /* The reason of the reallocation bug if memory is allocated dynamically using push_back() is as followed:
    The memory for the matrices are allocated continuously. If one of the dimension is increased, no such continuous memory space could be found, hence memory problem occurs.
    Alternatively, we can create a new matrix with the increased dimension and copy the entries of the old matrix to the new matrix, 
    but a lot of copying would be involved every time one of the dimensions is changed, hence we do not suggest it as a feasible method. */
	maxDimen = 30; //maximum dimension of the rho_j, povm_k or Cap_m matrix
	maxMatrix = 30; //maximum number of rho_j, povm_k or Cap_m to be used for computation
	
	//initialise rho, povm and Cap
	for(int s=0; s<maxDimen; s++) rho2.push_back(zero);
	for(int s=0; s<maxDimen; s++) rho1.push_back(rho2);
	for(int s=0; s<maxMatrix; s++) rho.push_back(rho1);
	
	for(int s=0; s<maxDimen; s++) povm2.push_back(zero);
	for(int s=0; s<maxDimen; s++) povm1.push_back(povm2);
	for(int s=0; s<maxMatrix; s++) povm.push_back(povm1);
	
	for(int s=0; s<maxDimen; s++) Cap3.push_back(zero);
	for(int s=0; s<maxDimen; s++) Cap2.push_back(Cap3);
	for(int s=0; s<maxDimen; s++) Cap1.push_back(Cap2);
	for(int s=0; s<maxMatrix; s++) Cap.push_back(Cap1);

	phi = 0.5+(0.5*sqrt(5.0));
}

//actual function that carry out the calculation of mutual information MI
void MutInf::calcMutualInformation()
{       
	//initialisation of variables
	int i, j, k, l, cmpPOVM = 0;
	dcomp ratio1, ratio2;
	dcomp temp(0,0);
	int restarted = 0;
	int counter = 0;
	double beta = 0.0;
	int beta_count = 0;
	int seed = (unsigned int)time(NULL);		
	new_MI = 0.0; MI = 0.0; alpha = 0.0; epsilon = 0.0;
	srand(seed);	

    //to determine the machine_epsilon
    double float_radix = 2.0;
    double inverse_radix = 1.0/float_radix;
    double machine_epsilon = 1.0;
    double temp_value = 1.0 + machine_epsilon;

    while(temp_value != 1.0) {
        machine_epsilon *= inverse_radix;
        temp_value = 1.0 + machine_epsilon;
    } //used later for the teimination criterion
    
	//initialise the output file
	ofstream outfile;
	outfile.open(output.c_str(), ios::out | ios::trunc);
	outfile << "Number of possible Quantum States J = " << J << "\n";
	outfile << "Number of POVM outcomes K = " << K << "\n";
	outfile << "Dimension of statistical operators N = " << N << "\n";
    outfile << "Number of Kraus Operators Cap_m = " << M << "\n";
    outfile << "Number of Channels Num_channels = " << C << "\n";
   	outfile << "Dimension of K_m = " << D << "\n";
	outfile << "Percentage chance to calculate with direct gradient = " << directGradChance << "/100 \n";
	outfile << "Tolerance in Mutual Information = " << tol_MI << "\n";
	outfile << "Starting seed for number generator = " << seed << "\n\n";
	
	//clear the probabilities stored previously
	prob_table.clear();
		
	vector< vector<dcomp> > A_dagger = Init2DMatrix(K,N);
	vector< vector<dcomp> > new_A_dagger = Init2DMatrix(K,N);
	vector< vector<dcomp> > S = Init2DMatrix(N,N);
	vector< vector<dcomp> > grad_MI = Init2DMatrix(K,N);
	vector< vector<dcomp> >	delta_A_dagger = Init2DMatrix(K,N);
	vector< vector<dcomp> > old_delta_A_dagger = Init2DMatrix(K,N);
	vector< vector<dcomp> > old_grad_MI = Init2DMatrix(K,N);
	old_grad_MI[0][0] = 1.0;
	vector< vector< vector<dcomp> > > R = Init3DMatrix(K,N,N);
    vector< vector< vector<dcomp> > > T = Init3DMatrix(J,N,N);
    vector< vector< vector<dcomp> > > init_rho = Init3DMatrix(J,N,N);
    vector< vector< vector<dcomp> > > new_rho;
	
	gen_A_dagger(A_dagger, S); //initialise A_dagger 
	gen_rho(rho); //initialize the input states rho
	init_rho = rho; //to print the intial generated input states rho;
	new_rho = rho;

restart:
		if (restarted == 1)
		{
			counter = 0;
			seed = seed + 1;
			srand(seed);
			gen_A_dagger(A_dagger,S);
			gen_rho(rho);
            init_rho = rho; 
            new_rho = rho;
		}
		restarted = 1;

        calc_prob_table(A_dagger, rho); //for the specified rho, calculate the probability table using A_dagger
        MI = calc_MI_from_prob_table(prob_table); //calculate the mutual information from the probability table
		if ((isnan(MI) || MI < 0.0)) goto restart;		
   			      	
		while (1)
		{
			counter++;
			if (counter>1000) goto restart;
			calc_R(R, rho);
			calc_grad_MI(grad_MI, R, A_dagger);
			
			//evaluate the chance to use direct gradient in the search for steepest ascent
			if((rand()%2)*100 < directGradChance){
				for (i=0; i<K; i++){
					for (j=0; j<N; j++) {
						delta_A_dagger[i][j] = grad_MI[i][j];
						old_delta_A_dagger[i][j] = grad_MI[i][j];
					}
				}				
			}else{
				calc_delta_A_dagger_CG(beta, beta_count, grad_MI, old_grad_MI, delta_A_dagger, old_delta_A_dagger);
			}
			//to find the improved rho and povm
			find_alpha_and_epsilon(new_A_dagger, A_dagger, delta_A_dagger, new_rho, rho);
			new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, rho);
			
			if ((isnan(new_MI) || new_MI < 0.0)) goto restart;
			outfile << "counter = " << counter << ", new_MI = " << new_MI << "\n";
			if (2.0*(new_MI - MI) <= tol_MI*(new_MI+MI)+machine_epsilon && counter > 20) break; //termination criteria
			swap2D(new_A_dagger, A_dagger);			
			swap3D(new_rho, rho);
			MI = new_MI;                           			
		}   
        outfile << "\n\n";        
            
/////////////////////////////////////////////////////////////////////
        //------------------------------------------------
	    //	print the Kraus Operators of the Capacitor		
	    //------------------------------------------------	
        outfile << "The Kraus Operators of the Capacitor are: \n";	
		for(k=0;k<M;k++){
            outfile << "Cap" << (k+1) << "={";
	        for(i=0;i<D;i++){
			    outfile << "{";
	            for(j=0;j<D;j++){
			        outfile << real(Cap[0][k][i][j]);
			        if(imag(Cap[0][k][i][j])>0)outfile << "+";
	                if(imag(Cap[0][k][i][j])!=0)outfile << imag(Cap[0][k][i][j]) << "I";
			        if(j!= D-1) outfile << ",";
		        }
		        outfile << "}";
		        if(i!= D-1) outfile << ",";
	        }
	        outfile << "}\n";
        }
        outfile << "\n";
////////////////////////////////////////////////////////////////////
        //---------------------------------------------
	    //	print the input states to file		
        //---------------------------------------------	
        outfile << "The initial input states are: \n";	
		for(k=0;k<J;k++){
            outfile << "" << (k+1) << "={";
	        for(i=0;i<N;i++){
		        outfile << "{";
		        for(j=0;j<N;j++){
			        outfile << real(init_rho[k][i][j]);
			        if(imag(init_rho[k][i][j])>0)outfile << "+";
			        if(imag(init_rho[k][i][j])!=0)outfile << imag(init_rho[k][i][j]) << "I";
			        if(j!= N-1) outfile << ",";
		        }
		        outfile << "}";
		        if(i!= N-1) outfile << ",";
	        }
	        outfile << "}\n";
        }
        outfile << "\n";
///////////////////////////////////////////////////////////////////
        outfile << "After optimization, the input states are: \n";	
        //--------------------------------------------
        //  print the optimized input states to file
        //--------------------------------------------
        for(k=0;k<J;k++){
	        outfile << "" << (k+1) << "={";
	        for(i=0;i<N;i++){
	        	outfile << "{";
		        for(j=0;j<N;j++){
			        outfile << real(rho[k][i][j]);
			        if(imag(rho[k][i][j])>0) outfile << "+";
			        if(imag(rho[k][i][j])!=0) outfile << imag(rho[k][i][j]) << "I";
			        if(j!= N-1) outfile << ",";
		        }
		        outfile << "}";
		        if(i!= N-1) outfile << ",";
	        }
		        outfile << "}\n";
        }
        outfile << "\n";
		//form the POVM before elimination
		for(k=0;k<K;k++){
			outer_product_add(povm[k],A_dagger[k],A_dagger[k]);
		} 
///////////////////////////////////////////////////////////////////
		//---------------------------------------------
		//	print the povm before elimination to file		
		//---------------------------------------------
		outfile << "The initial POVM outcomes generated are: \n";
		for(k=0;k<K;k++){
			outfile << "" << (k+1) << "={";
			for(i=0;i<N;i++){
				outfile << "{";
				for(j=0;j<N;j++){
					outfile << real(povm[k][i][j]);
					if(imag(povm[k][i][j])>0)outfile << "+";
					if(imag(povm[k][i][j])!=0)outfile << imag(povm[k][i][j]) << "I";
					if(j!= N-1) outfile << ",";
				}
				outfile << "}";
				if(i!= N-1) outfile << ",";
			}
			outfile << "}\n";			
		}
        outfile << "\n"; 
////////////////////////////////////////////////////////////////////				
		outfile << "After eliminating the extra POVM outcomes: \n";							
		//---------------------------------------------
		//  eliminate the extra povms
		//---------------------------------------------
		calc_prob_table(A_dagger, new_rho); 
		for(k=K-1;k>-1;k--){
			for(i=0;i<k;i++){
				double test,temp1=0,temp2=0,temp3=0;
				for(j=0;j<J;j++){
					temp1 += prob_table[j][k]*prob_table[j][i];
					temp2 += prob_table[j][k]*prob_table[j][k];
					temp3 += prob_table[j][i]*prob_table[j][i];
				}

				test = 1.0 - temp1*temp1/(temp2*temp3);
				//if found two POVM to be directly proportional to each other, add them up and assign to the ith POVM, move up the last POVM to the kth POVM being compared, reduced the total number of POVM, i.e. K
				if(test<0.001){
					matrix_add_matrix(povm[i],povm[k]);
					swap2D(povm[k],povm[K-1]);
					K--;
					break;
				}				
			}
		}
		//---------------------------------------------
		//	print the final povm to file		
		//---------------------------------------------
		for(k=0;k<K;k++){
			outfile << "" << (k+1) << "={";
			for(i=0;i<N;i++){
				outfile << "{";
				for(j=0;j<N;j++){
					outfile << real(povm[k][i][j]);
					if(imag(povm[k][i][j])>0)outfile << "+";
					if(imag(povm[k][i][j])!=0)outfile << imag(povm[k][i][j]) << "I";
					if(j!= N-1) outfile << ",";
				}
				outfile << "}";
				if(i!= N-1) outfile << ",";
			}
			outfile << "}\n";			
		} 
		//close the output file
		outfile.close();
}

//------------------------------------------------------------------------------
//	functions dealing with actual calculation
//------------------------------------------------------------------------------
//The calculated value of beta equals to gamma_i defined in Eq. (10.6.7) in p. 422 of NRC, or beta_k in Eq.(2.7.32) in p. 84 of NRC
double MutInf::calc_beta(vector< vector<dcomp> >& grad_MI, vector< vector<dcomp> >& old_grad_MI)
{
	double beta;
	int k;
	dcomp tempnum = 0.0;
	dcomp tempden = 0.0;
	for (k=0; k<K; k++) {
		tempnum += inner_product(grad_MI[k],grad_MI[k]) - inner_product(grad_MI[k],old_grad_MI[k]);
		tempden += inner_product(old_grad_MI[k],old_grad_MI[k]);
	}
	beta = real(tempnum)/real(tempden);
	return beta;
}

//This subroutine computes the conjugate gradient in a single iteration following the method in p. 83-86, or p. 420-423 of NRC
void MutInf::calc_delta_A_dagger_CG(double beta, int beta_count, vector< vector<dcomp> >& grad_MI, vector< vector<dcomp> >& old_grad_MI, vector< vector<dcomp> >& delta_A_dagger, vector< vector<dcomp> >& old_delta_A_dagger)
{
	/* Conjugate Gradient (CG) algorithm */
	int k, n;
	beta = calc_beta(grad_MI, old_grad_MI);
	dcomp tempsum = 0.0;
	if ( (beta_count < K*N*2) && (beta > 0) ) {
		for (k=0; k<K; k++){
			for (n=0; n<N; n++) {
                //this line correspond to Eq. (10.6.2) in NRC
				//delta_A_dagger corresponds to h_{i+1} in Eq. (10.6.2), or p_{k+1} in Eq. (2.7.32) in NRC
				//grad_MI corresponds to g_{i+1} in Eq. (10.6.2), or r_{k+1} in Eq. (2.7.32) in NRC
				delta_A_dagger[k][n] = grad_MI[k][n] + beta * old_delta_A_dagger[k][n];
			}
		}
		
		for (k=0;k<K;k++) {
			tempsum += inner_product(delta_A_dagger[k],grad_MI[k]);
		}
		//check that inner_product(delta_A_dagger, grad_MI) is positive
		if ( real(tempsum) > 0 ){
			beta_count++;
		}else{
			swap2D(delta_A_dagger, grad_MI);
			beta_count=0;
		}
	}
	else{
		swap2D(delta_A_dagger, grad_MI);
		beta_count=0;
	}

	//backup grad_MI to old_grad_MI; after that it doesn't matter what happens to grad_MI so we can just swap
	swap2D(grad_MI, old_grad_MI);

	//backup delta_A_dagger to old_delta_A_dagger; cannot swap because delta_A_dagger is required for calculating new_A_dagger 
	for (k=0; k<K; k++) {
		for (n=0; n<N; n++) {
			old_delta_A_dagger[k][n] = delta_A_dagger[k][n];
		}
	}
}

void MutInf::calc_grad_MI(vector< vector<dcomp> >& m, vector< vector< vector<dcomp> > >& a, vector< vector<dcomp> >& b)
{
	int i,j,k;
	vector< vector<dcomp> > GAMMA = Init2DMatrix(N,N);
	vector< vector<dcomp> > R_minus_GAMMA = Init2DMatrix(N,N);
	vector<dcomp> RA_dagger(N,0);
	
	for (k=0; k<K; k++) {
		matrix_times_vector(RA_dagger, a[k], b[k]);
		outer_product_add(GAMMA, RA_dagger, b[k]);
	}
	transpose_conjugate(GAMMA);

	for (k=0; k<K; k++) {
		//subtract GAMMA from R
		for(i=0;i<N;i++){
			for(j=0;j<N;j++){
				R_minus_GAMMA[i][j]=a[k][i][j] - GAMMA[i][j];
			}
		}
		matrix_times_vector(m[k], R_minus_GAMMA, b[k]);
	}
	//when this routine is executed till this line, grad_MI[k] = (R_k - \sum_l \pi_l R_l)A_dagger_k
}

void MutInf::calc_L_inverse(vector< vector<dcomp> >& L, vector< vector<dcomp> >& L_inv, vector< vector<dcomp> >& S)
{
	//calculate L where S = L x L_dagger, where L is a lower triangular matrix, then find L_inverse by back substitution
	//first calculate L
	//L_ij = 1/L_jj (S_ij - \sum_{k=0}^{j-1}L_ik L*_jk), i>j
	//L_ii = \sqrt{S_ii - \sum_{k=0}^{i-1} L_{ik}L*_{jk}}
	int row, col, k;
	dcomp sum;
	for (row=0; row<N; row++)
	{
		for (col=0; col<row+1; col++)
		{
			//first obtain summation \sum_{k=0}^{j-1}L_ik L*_jk
			sum = 0.0;
			for (k=0; k<col; k++)
			{
				sum = sum + L[row][k] * conj(L[col][k]);
			}

			if (col < row)
			{
				L[row][col] = (S[row][col] - sum) / conj(L[col][col]);
			}
			else
			{
				L[row][col] = sqrt(S[row][col] - sum);
			}
		}
	}

	//now calculate L_inverse, assuming that L_inverse is a lower triangular matrix as well
	//for i!=j, \sum_{k=j}^{k=i-1} L_ik L^{-1}_kj = - L_ii L^{-1}_ij
	for (col=0; col<N; col++)
	{
		L_inv[col][col] = 1.0 / L[col][col];
		for (row=col+1; row<N; row++)
		{
			sum = 0;
			for (k=col; k<row; k++)
			{
				sum = sum + L[row][k] * L_inv[k][col]; 
			}
			L_inv[row][col] = -sum / L[row][row];
		}
	}
}

inline double MutInf::calc_MI(double some_alpha, double some_epsilon, vector< vector<dcomp> >& new_A_dagger, vector< vector<dcomp> >& A_dagger, vector< vector<dcomp> >& delta_A_dagger, vector< vector< vector<dcomp> > >& opt_rho, vector< vector< vector<dcomp> > >& any_rho)
{
	vector< vector< vector<dcomp> > > T = Init3DMatrix(J, N, N);
    double temp_MI;
    
    calc_prob_table(A_dagger, any_rho);
    temp_MI = calc_MI_from_prob_table(prob_table);
    calc_T(T, A_dagger);
    calc_new_rho(some_epsilon, temp_MI, opt_rho, any_rho, T); //new input states rho

    calc_new_A_dagger(some_alpha, new_A_dagger, A_dagger, delta_A_dagger); //new povm
    calc_prob_table(new_A_dagger, opt_rho);
    return calc_MI_from_prob_table(prob_table);
}

double MutInf::calc_MI_from_prob_table(vector< vector<double> >& a)
{
	int j, k;
	double temp=0.0;
	for (j=0; j<J; j++)
	{
		for (k=0; k<K; k++)
		{
			temp = temp + a[j][k] * ( LOG2(a[j][k]) - LOG2(a[j][K]) - LOG2(a[J][k]) );
		}
	}
	return temp;
}

void MutInf::calc_new_A_dagger(double some_alpha, vector< vector<dcomp> >& new_A_dagger, vector< vector<dcomp> >& A_dagger, vector< vector<dcomp> >& delta_A_dagger)
{
	int k, n;
	vector< vector<dcomp> > S = Init2DMatrix(N,N);
	vector< vector<dcomp> > tmp_S_inverse_sqrt = Init2DMatrix(N,N);
	vector< vector<dcomp> > S_inverse_sqrt = Init2DMatrix(N,N);	
	vector< vector<dcomp> > tmp_A_dagger = Init2DMatrix(K,N);	
	
	for (k=0; k<K; k++) {
		for (n=0; n<N; n++) {
			new_A_dagger[k][n] = A_dagger[k][n] + some_alpha * delta_A_dagger[k][n];
		}
	}

	for (k=0; k<K; k++) {
		outer_product_add(S, new_A_dagger[k], new_A_dagger[k]);
	}
	
	//calc_S_inverse_sqrt()
	calc_L_inverse(tmp_S_inverse_sqrt, S_inverse_sqrt, S);

	for (k=0; k<K; k++) {
		matrix_times_vector(tmp_A_dagger[k], S_inverse_sqrt, new_A_dagger[k]);
		scalar_times_vector(new_A_dagger[k], 1.0, tmp_A_dagger[k]);
	}
}

//to calculate the new input states rho
void MutInf::calc_new_rho(double some_epsilon, double any_MI, vector< vector< vector<dcomp> > >& opt_rho, vector< vector< vector<dcomp> > >& any_rho, vector< vector< vector<dcomp> > >& T)
{
    dcomp sum_trace = (0.0, 0.0);
    double modulus = 0.0;
    int i, j;
    vector< vector< vector<dcomp> > > tmp_rho = Init3DMatrix(J,N,N);
    vector< vector< vector<dcomp> > > cor_rho = Init3DMatrix(J,N,N);
    vector< vector<dcomp> > ident1 = Init2DMatrix(N,N);
    vector< vector<dcomp> > ident2 = Init2DMatrix(N,N);
  
    for(i=0; i<N; i++){
        ident1[i][i] = 1.0;
        ident2[i][i] = -1.0;
    }
  
    scalar_times_matrix(ident2, any_MI, ident2);
    for(j=0; j<J; j++)
    {
        matrix_add_matrix(T[j], ident2);
        scalar_times_matrix(cor_rho[j], some_epsilon, T[j]);
        matrix_add_matrix(cor_rho[j], ident1);
        matrix_times_matrix(tmp_rho[j], any_rho[j], cor_rho[j]);
        transpose_conjugate(cor_rho[j]);
        matrix_times_matrix(opt_rho[j], cor_rho[j], tmp_rho[j]);
        sum_trace += cal_trace(opt_rho[j]);
    }
  
    modulus = cal_modulus(sum_trace) * cal_modulus(sum_trace);
  
    for(j=0; j<J; j++)
    {
        complex_times_matrix(opt_rho[j], conj(sum_trace), opt_rho[j]);
        scalar_times_matrix(opt_rho[j], 1.0/modulus, opt_rho[j]);
    }
  
}

void MutInf::calc_prob_table(vector< vector<dcomp> >& any_A_dagger, vector< vector< vector<dcomp> > >& any_rho)
{
    vector< vector< vector<dcomp> > > updated_rho =  Init3DMatrix(J, N, N);
	rho_after_Capacitor(updated_rho, any_rho);
	
	int j, k;
	double temp;
	vector< vector<dcomp> > tmp_A_dagger = Init2DMatrix(K,N);
	prob_table = Init2DRealMatrix(J+1,K+1);
	
	for (j=0; j<J; j++)
	{
		temp = 0.0;
		for (k=0; k<K; k++)
		{
			matrix_times_vector(tmp_A_dagger[k], updated_rho[j], any_A_dagger[k]);
			prob_table[j][k]=real(inner_product(any_A_dagger[k], tmp_A_dagger[k]));
			if (prob_table[j][k] < 0.0) prob_table[j][k]=0.0;
			temp += prob_table[j][k];
		}
		prob_table[j][K] = temp;
	}
	for (k=0; k<K; k++)
	{
		temp = 0.0;
		for (j=0; j<J; j++)
		{
			temp += prob_table[j][k];
		}
		prob_table[J][k] = temp;
	}
	
}

/*to calculate R_k*/
void MutInf::calc_R(vector< vector< vector<dcomp> > >& a, vector< vector< vector <dcomp> > >& any_rho) //R depends on rho, Cap and prob_table
{
    vector< vector< vector<dcomp> > > updated_rho =  Init3DMatrix(J, N, N);
	int  j, k, row, col;
	rho_after_Capacitor(updated_rho, any_rho);

	for (k=0; k<K; k++) {
		for (row=0; row<N; row++) {
			for (col=0; col<N; col++) {
				a[k][row][col]=0.0;
				for (j=0; j<J; j++) {
                     a[k][row][col] += updated_rho[j][row][col] * LOG2( prob_table[j][k] / (prob_table[j][K] * prob_table[J][k]) );
				}
			}
		}
	}
}

/*to calculate T_j*/
void MutInf::calc_T(vector< vector< vector<dcomp> > >& a, vector< vector<dcomp> >& A_dagger) //T depends on povm, Cap and prob_table
{
    vector< vector< vector<dcomp> > > new_povm =  Init3DMatrix(K, N, N);
	int j, k, row, col;
    povm_after_Capacitor(new_povm, A_dagger);
    
	for (j=0; j<J; j++) {
		for (row=0; row<N; row++) {
			for (col=0; col<N; col++) {
				  a[j][row][col]=0.0;
                  for (k=0; k<K; k++) {
			     		a[j][row][col] += new_povm[k][row][col] * LOG2( prob_table[j][k] / (prob_table[j][K] * prob_table[J][k]) );
				}
			}
		}
	}
}

// The computation of alpha and epsilon is carried out following the routine golden() in p. 401-402 of NRC. The idea of the algorithm is explained in section 10.2 from p. 397-400 NRC.
void MutInf::find_alpha_and_epsilon(vector< vector<dcomp> >& new_A_dagger, vector< vector<dcomp> >& A_dagger, vector< vector<dcomp> >& delta_A_dagger, vector< vector< vector<dcomp> > >& new_rho, vector< vector< vector<dcomp> > >& any_rho)
{
	double alphaR, alphaL, MIR, MIL, interval, current_max;
	alpha = 1.0;
    double epsilonR, epsilonL, tmp_MI;
    epsilon = 0.0;
	new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
	
    /* find the optimal value for alpha*/
	/* find an upper bound for the optimal alpha */
	if (new_MI > MI)
	{
		current_max = new_MI;
		while (1)
		{
			/* increase alpha until MI(alpha) < MI */
			alpha = alpha + 1.0;
			new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
			if (isnan(new_MI) || new_MI < current_max)
			{
				alpha = alpha - 1.0; break; //p. 399 in NRC
			}
			else current_max = new_MI;
			if (alpha > 1000.0) break; //limit alpha
		}
	}
	else
	{
		while (1)
		{
			/* half alpha until MI(alpha) > MI */
			alpha=alpha*0.5;
			if (alpha < tol_alpha) break;
			new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
			if (new_MI > MI) {alpha = alpha*2.0; break;}
		}
	}

	/* apply Golden Section Search */
	alphaR = phi * alpha;
	alphaL = (phi-1.0) * alpha;
	MIR = calc_MI(alphaR, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
	MIL = calc_MI(alphaL, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
	interval = alpha / (2.0*phi + 1.0);
	while (interval > tol_alpha * (alphaR + alphaL))
	{
		interval = interval * (phi-1.0);
		if (MIR > MIL)
		{
			alphaL = alphaR;
			MIL = MIR;
			alphaR = alphaR + interval;
			MIR = calc_MI(alphaR, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
		}
		else
		{
			alphaR = alphaL;
			MIR = MIL;
			alphaL = alphaL - interval;
			MIL = calc_MI(alphaL, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
		}
	}
	alpha = (alphaR + alphaL)*0.5;

        tmp_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
        epsilon = 1.0;
        new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
    
        /* find the optimal value for epsilon */
	/* find an upper bound for the optimal epsilon */
	if (new_MI > tmp_MI)
	{
		current_max = new_MI;
		while (1)
		{
			/* increase epsilon until MI(epsilon) < tmp_MI */
			epsilon = epsilon + 1.0;
			new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
			if (isnan(new_MI) || new_MI < current_max)
			{
				epsilon = epsilon - 1.0; break;
			}
			else current_max = new_MI;
			if (epsilon > 1000.0) break; //limit epsilon
		}
	}
	else
	{
		while (1)
		{
			/* half epsilon until MI(epsilon) > tmp_MI */
			epsilon = epsilon * 0.5;
			if (epsilon < tol_epsilon) break;
			new_MI = calc_MI(alpha, epsilon, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
			if (new_MI > tmp_MI) {epsilon = epsilon*2.0; break;}
		}
	} 

	/* apply Golden Section Search */
	epsilonR = phi * epsilon;
	epsilonL = (phi-1.0) * epsilon;
	MIR = calc_MI(alpha, epsilonR, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
	MIL = calc_MI(alpha, epsilonL, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
	interval = epsilon / (2.0*phi + 1.0);
	while (interval > tol_epsilon * (epsilonR + epsilonL))
	{
		interval = interval * (phi-1.0);
		if (MIR > MIL)
		{
			epsilonL = epsilonR;
			MIL = MIR;
			epsilonR = epsilonR + interval;
			MIR = calc_MI(alpha, epsilonR, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
		}
		else
		{
			epsilonR = epsilonL;
			MIR = MIL;
			epsilonL = epsilonL - interval;
			MIL = calc_MI(alpha, epsilonL, new_A_dagger, A_dagger, delta_A_dagger, new_rho, any_rho);
		}
	}
	epsilon = (epsilonR + epsilonL)*0.5;
}

//to generate the POVM randomly
void MutInf::gen_A_dagger(vector< vector<dcomp> >& A_dagger, vector< vector<dcomp> >& S)
{
	int i,j,k, n;
	vector< vector<dcomp> > tmp_A_dagger = Init2DMatrix(K,N);
	vector< vector<dcomp> > tmp_S_inverse_sqrt = Init2DMatrix(N,N);
	vector< vector<dcomp> > S_inverse_sqrt = Init2DMatrix(N,N);
	
	for (n=0; n<N; n++) {
		A_dagger[n][n] = 1.0;
	}
	
	for (k=0; k<K; k++) {
		for (n=0; n<N; n++) {
			dcomp random = complex<double>(0.002*((rand()/(double)RAND_MAX) - 0.5), 0.002*((rand()/(double)RAND_MAX)-0.5));
			A_dagger[k][n] = A_dagger[k][n] + random;
		}
	}
	
	for(i=0;i<N;i++){
		for(j=0;j<N;j++){
			S[i][j]=0.0;
		}
	}

	int test=0;
	for (k=0; k<K; k++) {
		outer_product_add(S, A_dagger[k], A_dagger[k]);
	}
	
	calc_L_inverse(tmp_S_inverse_sqrt, S_inverse_sqrt, S);

	for (k=0; k<K; k++) {
		matrix_times_vector(tmp_A_dagger[k], S_inverse_sqrt, A_dagger[k]);
		scalar_times_vector(A_dagger[k], 1.0, tmp_A_dagger[k]);
	}	
}

//to generate the input states rho randomly
void MutInf::gen_rho(vector< vector< vector<dcomp> > >& any_rho)
{
	int j, n;
	double sum_trace = 0.0;
	
	for (j=0; j<J; j++) {
		for (n=0; n<N; n++) {
			double random = rand()/(double)RAND_MAX - 0.5;
			any_rho[j][n][n] = 1.0 + random;
		}
		sum_trace += real(cal_trace(any_rho[j]));
	}
	
	for (j=0; j<J; j++) {
        scalar_times_matrix(any_rho[j], 1.0/sum_trace, any_rho[j]);
    }
}

void MutInf::rho_after_Capacitor(vector< vector< vector<dcomp> > >& rho_after, vector< vector< vector<dcomp> > >& any_rho)
{       
        int num = cal_power(M, C);
        vector< vector< vector<dcomp> > > temp_rho = Init3DMatrix(J, N, N);
        vector< vector< vector<dcomp> > > comp_rho = Init3DMatrix(J, N, N);
        vector< vector< vector<dcomp> > > Cap_after = Init3DMatrix(num, N, N);
        
        cal_kroneckerProduct(Cap_after, Cap[0]);
        
        for (int j=0; j<J; j++) {
           for (int k=0; k<num; k++) {  
              matrix_times_matrix(temp_rho[j], Cap_after[k], any_rho[j]);
              transpose_conjugate(Cap_after[k]);
              matrix_times_matrix(comp_rho[j], temp_rho[j], Cap_after[k]);
              matrix_add_matrix(rho_after[j], comp_rho[j]);
           }
        }
}

void MutInf::povm_after_Capacitor(vector< vector< vector<dcomp> > >& povm_after, vector< vector<dcomp> >& A_dagger)
{        
        int num = cal_power(M, C);
        vector< vector< vector<dcomp> > > any_povm = Init3DMatrix(K, N, N);
        vector< vector< vector<dcomp> > > temp_povm = Init3DMatrix(K, N, N);
        vector< vector< vector<dcomp> > > comp_povm = Init3DMatrix(K, N, N);
        vector< vector< vector<dcomp> > > Cap_after = Init3DMatrix(num, N, N);
        
        cal_kroneckerProduct(Cap_after, Cap[0]);
        
        for (int k=0; k<K; k++) {
           outer_product_add(any_povm[k], A_dagger[k], A_dagger[k]);
           for (int j=0; j<num; j++) {  
              matrix_times_matrix(temp_povm[k], Cap_after[j], any_povm[k]);
              transpose_conjugate(Cap_after[j]);
              matrix_times_matrix(comp_povm[k], temp_povm[k], Cap_after[j]);
              matrix_add_matrix(povm_after[k], comp_povm[k]);
           }
        }
}

//------------------------------------------------------------------------------
//	functions for interactions with GUI
//------------------------------------------------------------------------------
int MutInf::getK()
{
	return K;
}

int MutInf::getJ()
{
	return J;
}

int MutInf::getM()
{
	return M;
}

int MutInf::getC()
{
	return C;
}

vector< vector<dcomp> > MutInf::getPOVM(int X)
{
	return povm[X];
}

vector< vector<dcomp> > MutInf::getrho(int X)
{
	return rho[X];
}

vector< vector<dcomp> > MutInf::getCap(int X, int Y)
{
	return Cap[X][Y];
}

void MutInf::setrhoElement(int x, int y, int z, double d, int ri)
{
	if(ri == 0){
		dcomp temp(d,imag(rho[x][y][z]));
		rho[x][y][z] = temp;
	}else{
		dcomp temp(real(rho[x][y][z]),d);
		rho[x][y][z] = temp;
	}
}

void MutInf::setPOVMElement(int x, int y, int z, double d, int ri)
{
	if(ri == 0){
		dcomp temp(d,imag(povm[x][y][z]));
		povm[x][y][z] = temp;
	}else{
		dcomp temp(real(povm[x][y][z]),d);
		povm[x][y][z] = temp;
	}
}

void MutInf::setCapElement(int l, int x, int y, int z, double h, int ri)
{
	if(ri == 0){
		dcomp temp(h,imag(Cap[l][x][y][z]));
		Cap[l][x][y][z] = temp;
	}else{
		dcomp temp(real(Cap[l][x][y][z]),h);
		Cap[l][x][y][z] = temp;
	}
}

void MutInf::UpdaterhoDimension(int j, int n)
{
	int x,y,z;
	for(x=0;x<maxMatrix;x++){
		for(y=0;y<maxDimen;y++){
			for(z=0;z<maxDimen;z++){
				if(!(x<j && y<n && z<n))rho[x][y][z] = zero;
			}
		}
	}
	N = n;
	J = j;
}

void MutInf::UpdatePOVMDimension(int k, int n)
{
	int x,y,z;
    for(x=0;x<maxMatrix;x++){
		for(y=0;y<maxDimen;y++){                            
			for(z=0;z<maxDimen;z++){                                              
				povm[x][y][z]=zero;
			}
		}
	}		
	if(k<n){
		K = n;
	}else{
		K = k;
	}
	povmDimension = n;
}

void MutInf::UpdateCapDimension(int c, int m, int d)
{
	int l,x,y,z;
	for(l=0;l<maxMatrix;l++)
	for(x=0;x<maxMatrix;x++){
		for(y=0;y<maxDimen;y++){
			for(z=0;z<maxDimen;z++){
				if(!(l<c && x<m && y<d && z<d)) Cap[l][x][y][z]=zero;
			}
		}
	}
	D = d;
	C = c;
	M = m;
}

//-------------------------------------------------------------------------------------
//	functions to initialize 1D, 2D and 3D zero matrices, with all entries set to 0
//-------------------------------------------------------------------------------------
vector<dcomp> MutInf::InitVector(int X)
{
	vector<dcomp> a(X,0); 
	
	return a;
}

vector< vector<dcomp> > MutInf::Init2DMatrix(int X, int Y)
{
	vector<dcomp> b(Y,0); /* 2nd dimension */
	vector< vector<dcomp> > a(X,b);
	
	return a;
}

vector< vector<double> > MutInf::Init2DRealMatrix(int X, int Y)
{
	vector<double> b(Y,0); /* 2nd dimension */
	vector< vector<double> > a(X,b);
	
	return a;
}

vector< vector< vector<dcomp> > > MutInf::Init3DMatrix(int X, int Y, int Z)
{
	vector<dcomp> c(Z,0); /* 3rd dimension */
	vector< vector<dcomp> > b(Y,c); /* 2nd dimension */
	vector< vector< vector<dcomp> > > a(X,b);
	
	return a;
}

vector< vector< vector< vector<dcomp> > > > MutInf::Init4DMatrix(int X, int Y, int Z, int D)
{
	vector<dcomp> d(D,0); /* 4th dimension */
	vector< vector<dcomp> > c(Z,d); /* 3rd dimension */
	vector< vector< vector<dcomp> > > b(Y,c); /* 2nd dimension */
	vector< vector< vector< vector<dcomp> > > > a(X,b);
	
	return a;
}

//------------------------------------------------------------------------------
//	functions to multiply scalar, vectors and matrices
//------------------------------------------------------------------------------

dcomp MutInf::inner_product(vector<dcomp>& v1, vector<dcomp>& v2)
{
	int i;
	dcomp temp = 0.0;
	for (i=0; i<N; i++)
	{
		temp += conj(v1[i]) * v2[i];
	}
	return temp;
}

double MutInf::LOG2(double x)
{
	if (x==0.0) return 0.0;
	return log(x)/log(2.0);
}

//add two N by N matrix m2 and m3 and assign to m2;
void MutInf::matrix_add_matrix(vector< vector<dcomp> >& m2, vector< vector<dcomp> >& m3)
{
	int i,j;
	for (i=0; i<N; i++)	
	{
		for (j=0; j<N; j++)	
		m2[i][j] += m3[i][j];		
	}
}

//multiply two N by N matrix m2 and m3 and assign the product to m1
void MutInf::matrix_times_matrix(vector< vector<dcomp> >& m1, vector< vector<dcomp> >& m2, vector< vector<dcomp> >& m3)
{
	int i, j, k;
	for (i=0; i<N; i++)
	{
		for (k=0; k<N; k++)
		{
			m1[i][k] = 0;
			for (j=0; j<N; j++)
			{
				m1[i][k] += m2[i][j] * m3[j][k];
			}
		}
	}
}

//multiply an N by N matrix m with an N-element vector v2 and assign the product to v1
void MutInf::matrix_times_vector(vector<dcomp>& v1, vector< vector<dcomp> >& m, vector<dcomp>& v2)
{
	int i, j;
	for (i=0; i<N; i++)
	{
		v1[i] = 0;
		for (j=0; j<N; j++)
		{
			v1[i] += m[i][j] * v2[j];
		}
	}
}

void MutInf::outer_product_add(vector< vector<dcomp> >& m, vector<dcomp>& v1, vector<dcomp>& v2)
{
	int row, col;
	dcomp temp;
	for (col=0; col<N; col++)
	{
		temp = conj(v2[col]);
		for (row=0; row<N; row++)
			m[row][col] += v1[row] * temp;
	}
}

//multiply an N-element vector v2 with a scalar s and assign the product to an N-element vector v1
void MutInf::scalar_times_vector(vector<dcomp>& v1, double s, vector<dcomp>& v2)
{
	int i;
	for (i=0; i<N; i++)
	{
		v1[i] = s * v2[i];
	}
}

//multiply an N by N matrix m2 with a scalar s and assign the product to m1
void MutInf::scalar_times_matrix(vector< vector<dcomp> >& m1, double s, vector< vector<dcomp> >& m2)
{
	int i, j;
	for (i=0; i<N; i++){
	    for(j=0; j<N; j++){
	       m1[i][j] = s * m2[i][j];
            }
	}
}

//multiply an N by N matrix m2 with a complex number s and assign the product to m1
void MutInf::complex_times_matrix(vector< vector<dcomp> >& m1, dcomp s, vector< vector<dcomp> >& m2)
{
	int i, j;
	for (i=0; i<N; i++){
	    for(j=0; j<N; j++){
	       m1[i][j] = s * m2[i][j];
            }
	}
}

void MutInf::swap2D(vector< vector<dcomp> >& m1, vector< vector<dcomp> >& m2)
{
	m1.swap(m2);
}

void MutInf::swap3D(vector< vector< vector<dcomp> > >& m1, vector< vector< vector<dcomp> > >& m2)
{
    int n = (int)m1[0].size();
    for(int i=0; i<n; i++){
       swap2D(m1[i], m2[i]);
    }
}

void MutInf::transpose_conjugate(vector< vector<dcomp> >& m)
{
	int row, col;
	dcomp temp;
	for (row=0; row<N; row++) {
		for (col=0; col <= row; col++) {
			temp = m[row][col];
			m[row][col] = conj(m[col][row]);
			m[col][row] = conj(temp);
		}
	}
}

//to calculate the trace of a square matrix
dcomp MutInf::cal_trace(vector< vector<dcomp> >& m)
{
    dcomp trace = 0.0;
    for(int i=0; i<N; i++){
       trace += m[i][i];
    }
    return trace; 
}

//to calculate the modulus of a complex number
double MutInf::cal_modulus(dcomp x)
{
     double modulus = 0.0;
     return modulus = sqrt(real(x)*real(x) + imag(x)*imag(x));
}

//to calculate the kronecker products of C matrices
void MutInf::cal_kroneckerProduct(vector< vector< vector<dcomp> > >& m, vector< vector< vector<dcomp> > >& m1)
{
    int l;
    int dim1, dim2, num1, num2;
    
    for (l=C; l>0; l--) {
       dim1 = cal_power(D, C-l);
       dim2 = cal_power(D, C-l+1);
       num1 = cal_power(M, C-l); 
       num2 = cal_power(M, C-l+1);
       
       if (l==C) m = m1;
       else {
           vector< vector< vector<dcomp> > > temp = Init3DMatrix(num1, dim1, dim1);
           temp = m;
           vector< vector< vector<dcomp> > > m = Init3DMatrix(num2, dim2, dim2);
           
           for (int i=0; i<num2; i++) { //number of elements in new m
              for (int j=0; j<num1; j++) { //number of elements in old m
                 for (int k=0; k<M; k++) { //number of elements in m1, i.e. M not changed
                    for (int p=0; p<dim2; p++) {
                       for (int q=0; q<dim2; q++) {
                          m[i][p][q] = temp[j][p/dim1][q/dim1]*m1[k][p%D][q%D];
                       }
                    }
                 }
              }
           } 
       }
    }
}

//to calculate a to the power of b
int MutInf::cal_power(int a, int b)
{
    int c = a;
    for (int i=b; i>1; i--) c *= a;
    return c;
}
